import logging
import sys
from typing import Callable, Dict, List, Mapping, Type

import click
from ruamel.yaml.parser import ParserError
from ruamel.yaml.scanner import ScannerError

from forge_template.exception.exceptions import AccessDeniedException, ValidationException
from forge_template.exception.handler import (
    handle_access_denied_error,
    handle_file_not_found,
    handle_parser_error,
    handle_validation_error,
)
from forge_template.handler.azure_handler import AzureHandler
from forge_template.handler.databricks_handler import DatabricksHandler
from forge_template.handler.github_actions_handler import GithubActionsHandler
from forge_template.handler.handler import BaseHandler
from forge_template.handler.powerbi_handler import PowerBIHandler
from forge_template.util.config_generator import ConfigGenerator
from forge_template.util.user_interface import UserInterface
from forge_template.util.yaml_util import (
    AZURE_CONFIG_NAME,
    DATABRICKS_CONFIG_NAME,
    POWERBI_CONFIG_NAME,
    YAML_CONFIG_PATH,
    YAML_CONFIG_SCHEMA_PATH,
    YAML_SECRETS_PATH,
    YAML_SECRETS_SCHEMA_PATH,
    assert_same_services,
    assert_scope_matches,
    load_and_validate_yaml,
    merge_dictionaries,
)

HANDLER_CLASSES: Mapping[str, Type[BaseHandler]] = {
    DATABRICKS_CONFIG_NAME: DatabricksHandler,
    AZURE_CONFIG_NAME: AzureHandler,
    POWERBI_CONFIG_NAME: PowerBIHandler,
}
LOGGING_FILE = "logs.log"


def instantiate_config() -> Dict:
    # Parse configs & secrets:
    config = load_and_validate_yaml(YAML_CONFIG_PATH, YAML_CONFIG_SCHEMA_PATH)
    secrets = load_and_validate_yaml(YAML_SECRETS_PATH, YAML_SECRETS_SCHEMA_PATH)

    # Assert that config files have the same services defined
    assert_same_services(config, secrets)

    # Assert that scopes matches
    assert_scope_matches(config, secrets)

    # Merge config files:
    merge_dictionaries(config, secrets)

    return config


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    logging.basicConfig(filename="logs.log", level=logging.INFO)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))
    UserInterface.print_title("Forge Template")
    logging.info(
        "If you have any questions that cannot be answered by the readme, please contact magnus.moan@cognite.com\n"
    )

    if ctx.invoked_subcommand is None:
        show_main_menu()
        UserInterface.print_title("Good bye!")


@cli.command()
@click.option("--project-name", "-n", prompt=True, help="name of project")
@click.option("--tool", "-t", type=click.Choice(HANDLER_CLASSES.keys(), case_sensitive=False))
def generate(project_name: str = "", tool: str = "") -> None:
    """ Generate a new config file """
    run_generate(project_name, tool)


@cli.command()
@click.option("--tool", "-t", type=click.Choice(HANDLER_CLASSES.keys(), case_sensitive=False))
def deploy(tool: str) -> None:
    """ Deploy configuration """
    run_deploy(tool)


@cli.command()
@click.option("--tool", "-t", type=click.Choice(HANDLER_CLASSES.keys(), case_sensitive=False))
def delete(tool: str = "") -> None:
    """ Delete resources mentioned in configuration """
    run_delete(tool)


@cli.command()
def upload_config() -> None:
    """ Upload config.yaml to Databricks """
    run_upload_config()


@cli.command()
@click.option("--tool", "-t", type=click.Choice(HANDLER_CLASSES.keys(), case_sensitive=False))
def pipeline(tool: str) -> None:
    """ Set up pipelines for automatic deployment """
    run_create_pipeline(tool)


@cli.resultcallback()
def process_result(_, __):
    UserInterface.print_title("Good bye!")


def run_generate(project_name: str = "", tool: str = ""):
    generator = ConfigGenerator(project_name)

    def generate_action(chosen_tool: str) -> None:
        is_custom = click.confirm(
            f"Do you want to do a custom configuration of {chosen_tool} (only recommended in very special cases)?",
            default=False,
            show_default=True,
        )
        generator.generate(chosen_tool, is_custom=is_custom)
        generator.preview()
        logging.info("\n")
        if click.confirm("Do you want to save the preview?"):
            generator.save()

    show_sub_menu("generate config for", generate_action, tool)


def run_deploy(tool: str = "") -> None:
    def deploy_action(chosen_tool: str) -> None:
        config = instantiate_config()
        handler = HANDLER_CLASSES[chosen_tool](chosen_tool, config)
        handler.run_create_preview()
        handler.run_print_preview()
        if click.confirm("Do you want to proceed?", default=False):
            handler.run_setup()
            UserInterface.print_ok(f"Successfully deployed {chosen_tool}")

    show_sub_menu("deploy", deploy_action, tool)


def run_delete(tool: str = "") -> None:
    def delete_action(chosen_tool: str) -> None:
        config = instantiate_config()
        handler = HANDLER_CLASSES[chosen_tool](chosen_tool, config)
        handler.run_delete_all_resources()
        UserInterface.print_ok(f"Successfully deleted {chosen_tool}")

    show_sub_menu("delete", delete_action, tool)


def run_upload_config() -> None:
    def upload_config_action() -> None:
        config = instantiate_config()
        handler = DatabricksHandler(DATABRICKS_CONFIG_NAME, config)
        handler.upload_config()

    do_action(upload_config_action)
    show_main_menu()


def run_create_pipeline(tool: str = "") -> None:
    def create_pipeline_action(chosen_tool: str) -> None:
        handler = GithubActionsHandler(chosen_tool)
        handler.print_preview()
        if click.confirm("Do you want to proceed?", default=False):
            handler.run_setup()
            UserInterface.print_ok(
                f"Successfully set up pipeline (still needs to be pushed to git) for {chosen_tool}. "
                f"In order for the pipeline to work, the {chosen_tool.upper()}_TOKEN "
                f"needs to be set in the Github repository secrets"
            )

    show_sub_menu("set up a deployment pipeline for", create_pipeline_action, tool)


def show_main_menu():
    exit_cmd = "e"
    generate_cmd = "Generate config"
    deploy_cmd = "Deploy config"
    delete_cmd = "Delete resources"
    upload_cmd = "Upload config.yaml to Databricks"
    pipeline_cmd = "Setup deployment pipeline"

    choice_to_action = {
        generate_cmd: run_generate,
        deploy_cmd: run_deploy,
        delete_cmd: run_delete,
        upload_cmd: run_upload_config,
        pipeline_cmd: run_create_pipeline,
    }

    ui = UserInterface()
    prompt = "What do you want to do?"
    choice = ui.get_menu_selection(prompt, list(choice_to_action.keys()), is_top_level=True)
    if choice != exit_cmd:
        choice_to_action[choice]()


def prompt_for_option(action: str, options: List[str]) -> str:
    ui = UserInterface()
    prompt = f"Which tool do you want to {action}?"
    return ui.get_menu_selection(prompt, options)


def show_sub_menu(action_str: str, action_handler: Callable[[str], None], user_option: str = "") -> None:
    exit_command = "e"
    menu_command = "m"
    tools = list(map(lambda x: x.capitalize(), HANDLER_CLASSES.keys()))

    if not user_option:
        user_option = prompt_for_option(action_str, tools)
    if user_option not in [menu_command, exit_command]:
        do_action(lambda: action_handler(user_option.lower()))
    if user_option != exit_command:
        show_main_menu()


def do_action(action: Callable[[], None]) -> None:
    try:
        action()
    except FileNotFoundError as e:
        handle_file_not_found(e)
    except (ParserError, ScannerError) as e:
        handle_parser_error(e)
    except ValidationException as e:
        handle_validation_error(e)
    except AccessDeniedException as e:
        handle_access_denied_error(e)



from ruamel.yaml.parser import ParserError

from forge_template.exception.exceptions import AccessDeniedException, ValidationException
from forge_template.util.user_interface import UserInterface
from forge_template.util.yaml_util import YAML_CONFIG_PATH, YAML_SECRETS_PATH


def handle_file_not_found(e: FileNotFoundError) -> None:
    if e.filename == YAML_CONFIG_PATH or e.filename == YAML_SECRETS_PATH:
        UserInterface.print_failure(f"\nConfiguration file named {e.filename} not found. Please generate config.")
    else:
        raise e


def handle_parser_error(e: ParserError) -> None:
    UserInterface.print_failure(f"\n{str(e)}\nPlease verify that the content of your .yaml files is parseable")


def handle_validation_error(e: ValidationException) -> None:
    UserInterface.print_failure(f"\n{str(e)}")


def handle_access_denied_error(e: AccessDeniedException) -> None:
    UserInterface.print_failure(f"\n{str(e)}")



import os
import shutil
from typing import List

import forge_template.util.log_utils as utils
from forge_template.handler.handler import BaseHandler
from forge_template.util.paths import (
    DATABRICKS_SCRIPT_NAME,
    DATABRICKS_WORKFLOW_TEMPLATE_NAME,
    POWERBI_SCRIPT_NAME,
    POWERBI_WORKFLOW_TEMPLATE_NAME,
    SCRIPT_TEMPLATE_FOLDER,
    WORKFLOW_TEMPLATE_FOLDER,
)
from forge_template.util.yaml_util import DATABRICKS_CONFIG_NAME, POWERBI_CONFIG_NAME

SCRIPT_FOLDER = "pipeline_scripts"
WORKFLOW_ACTION_FOLDER = os.path.join(".github", "workflows")
GITHUB_ACTION_HANDLER_NAME = "github_actions"


class GithubActionsHandler(BaseHandler):
    def __init__(self, tool_name: str) -> None:
        self._tool = tool_name
        self._paths_to_create = [WORKFLOW_ACTION_FOLDER, SCRIPT_FOLDER]

        if tool_name == DATABRICKS_CONFIG_NAME:
            self._script_to_add = os.path.join(SCRIPT_FOLDER, DATABRICKS_SCRIPT_NAME)
            self._template_to_add = os.path.join(WORKFLOW_ACTION_FOLDER, DATABRICKS_WORKFLOW_TEMPLATE_NAME)
        elif tool_name == POWERBI_CONFIG_NAME:
            self._script_to_add = os.path.join(SCRIPT_FOLDER, POWERBI_SCRIPT_NAME)
            self._template_to_add = os.path.join(WORKFLOW_ACTION_FOLDER, POWERBI_WORKFLOW_TEMPLATE_NAME)
        else:
            raise RuntimeError(
                f"Unexpected tool. Got {self._tool}, but expected one of "
                f"{[POWERBI_CONFIG_NAME, DATABRICKS_CONFIG_NAME]}"
            )

        self._to_from_paths = [
            (self._script_to_add, os.path.join(SCRIPT_TEMPLATE_FOLDER, os.path.basename(self._script_to_add))),
            (self._template_to_add, os.path.join(WORKFLOW_TEMPLATE_FOLDER, os.path.basename(self._template_to_add))),
        ]

        self._created_folders: List[str] = []
        self._copied_files: List[str] = []

        super().__init__(name=GITHUB_ACTION_HANDLER_NAME, config={})

    def create_preview(self) -> None:
        pass

    def print_preview(self) -> None:
        utils.print_resource_to_add(self._paths_to_create, "Directory")
        utils.print_resource_to_add(
            [self._script_to_add, self._template_to_add], "File",
        )

    def setup(self) -> None:
        self.__create_path()
        self.__copy_files()

    def rollback(self) -> None:
        if self._copied_files:
            utils.print_rollback("Files")
            for file in self._copied_files:
                self.__delete_path(file)

        if self._created_folders:
            utils.print_rollback("Folders")
            for folder in self._created_folders:
                self.__delete_path(folder)

    def delete_all_resources(self):
        for path in self._paths_to_create:
            self.__delete_path(path)

    @staticmethod
    def __delete_path(path: str) -> None:
        if os.path.exists(path):
            if os.path.isdir(path):
                os.rmdir(path)
                utils.print_success_deleted(path, "Directory")
            else:
                os.remove(path)
                utils.print_success_deleted(path, "File")

    def __create_path(self) -> None:
        for path in self._paths_to_create:
            if not os.path.exists(path):
                os.makedirs(path)
                self._created_folders.append(path)
                utils.print_success_created(path, "Directory")

    def __copy_files(self) -> None:
        for (to_path, from_path) in self._to_from_paths:
            shutil.copyfile(from_path, to_path)
            self._copied_files.append(to_path)
            file_name = os.path.basename(to_path)
            folder_name = os.path.dirname(to_path)
            utils.print_success_added(folder_name, "Directory", file_name, "File")

    @property
    def created_folders(self):
        return self._created_folders



import os
import re
import sys
from typing import Any, Callable, Dict, List, Union

import click
from cerberus import Validator
from ruamel.yaml import YAML

from forge_template.util.user_interface import UserInterface
from forge_template.util.yaml_util import (
    YAML_CONFIG_PATH,
    YAML_CONFIG_SCHEMA_PATH,
    YAML_SECRETS_PATH,
    YAML_SECRETS_SCHEMA_PATH,
    load_yaml,
)


class ConfigGenerator:
    def __init__(self, project_name: str = ""):
        while not re.match(r"^[-\w_]+$", project_name):
            print("\nPlease enter a project name (string containing only letters, numbers, '_' and '-')")
            project_name = click.prompt("Project name", type=click.STRING)

        self.project_name = project_name
        self.ui = UserInterface()
        self.config: Dict[str, Any] = {}
        self.secrets: Dict[str, Any] = {}
        self.config_schema = load_yaml(YAML_CONFIG_SCHEMA_PATH)
        self.secrets_schema = load_yaml(YAML_SECRETS_SCHEMA_PATH)

        if os.path.exists(YAML_CONFIG_PATH):
            self.config = load_yaml(YAML_CONFIG_PATH)

        if os.path.exists(YAML_SECRETS_PATH):
            self.secrets = load_yaml(YAML_SECRETS_PATH)

    def generate(self, tool: str, is_custom: bool = False) -> None:
        self.config[tool] = self.__generate_config(tool, self.config_schema[tool], is_custom=is_custom)
        self.secrets[tool] = self.__generate_config(tool, self.secrets_schema[tool], is_custom=is_custom)

    def preview(self) -> None:
        yaml = YAML()
        print()
        self.ui.print_ok("Preview: config.yaml")
        yaml.dump(self.config, sys.stdout)

        print()
        self.ui.print_ok("Preview: secrets.yaml")
        yaml.dump(self.secrets, sys.stdout)

    def save(self) -> None:
        yaml = YAML()
        print()
        with open(YAML_CONFIG_PATH, "w") as f:
            yaml.dump(self.config, f)
            self.ui.print_ok("Saved config.yaml")

        with open(YAML_SECRETS_PATH, "w") as f:
            yaml.dump(self.secrets, f)
            self.ui.print_ok("Saved secrets.yaml")

    @staticmethod
    def __get_current_tool_config(tool: str, config: Dict[str, Any]) -> Dict[str, Any]:
        return config[tool] if tool in config.keys() else {}

    def __generate_config(
        self, key: str, schema: Dict[str, Any], is_custom: bool = False
    ) -> Union[Dict[str, Any], str, List[str]]:
        config = {}

        if type(schema) is dict:
            if schema["type"] == "dict":
                for key, values in schema["schema"].items():
                    config[key] = self.__generate_config(key, values, is_custom)

            elif schema["type"] == "list" and schema["schema"]["type"] == "dict":
                keys = list(dict(schema["schema"]["schema"]).keys())
                return self.ui.get_input_list_of_dicts(
                    schema["meta"]["prompt"],
                    validators=self.__get_list_of_dict_validators(schema["schema"]["schema"]),
                    keys=keys,
                )
            else:
                default_value = self.__get_default_value(key, schema, self.project_name)
                if not is_custom and default_value is not None:
                    return default_value
                else:
                    transform = self.__get_transformer(schema)
                    user_input = self.ui.get_input(
                        schema["meta"]["prompt"],
                        default=default_value if default_value else "",
                        validate=self.__validate(key, schema, transform),
                    )
                    return transform(user_input)

        return config

    def __validate(
        self, key: str, schema: Dict[str, Any], transform: Callable[[str], Any] = None
    ) -> Callable[[str], bool]:
        v = Validator({key: schema})

        def check_regex(input_value: str) -> bool:
            if transform:
                input_value = transform(input_value)
            is_valid = v.validate({key: input_value})
            if not is_valid:
                if "meta" in schema and "help" in schema["meta"]:
                    self.ui.print_warning(f"Explanation: {schema['meta']['help']}")
                self.ui.print_warning(f"Error message: {v.errors}")
            return is_valid

        return check_regex

    def __get_transformer(self, schema: Dict[str, Any]) -> Callable[[str], Any]:
        if schema["type"] == "list":
            return self.__transform_list
        else:
            return self.__base_transform

    def __get_list_of_dict_validators(self, schema: Dict[str, Any]) -> Dict[str, Callable[[str], bool]]:
        validators = {}
        for key in schema:
            validators[key] = self.__validate(key, schema[key])

        return validators

    @staticmethod
    def __transform_list(input_value: str) -> List[str]:
        return list(map(lambda x: x.strip(), input_value.split(",")))

    @staticmethod
    def __base_transform(input_value: str) -> str:
        return input_value.strip()

    @staticmethod
    def __get_default_value(key: str, schema: Dict[str, Any], project_name: str) -> Union[str, None]:
        if "default" in schema:
            value = schema["default"]
            if "<PROJECT_NAME>" in value:
                default_value = value.replace("<PROJECT_NAME>", project_name)
                if key == "database_name":
                    default_value = default_value.replace("-", "_").replace(" ", "_")
                return default_value
            return value
        return None



import fnmatch
import os
import sys

from pbiapi import PowerBIAPIClient
from ruamel.yaml import YAML

CONFIG_FILE_PATH = "config.yaml"
POWERBI_FILES_FOLDER = "powerbi"
POWERBI_DATASET_FOLDER = os.path.join(POWERBI_FILES_FOLDER, "datasets")
POWERBI_REPORTS_FOLDER = os.path.join(POWERBI_FILES_FOLDER, "reports")


def get_display_name(path: str) -> str:
    return os.path.splitext(os.path.basename(path))[0]


with open(CONFIG_FILE_PATH) as f:
    config = YAML(typ="safe").load(f)["powerbi"]

if len(sys.argv) < 2:
    raise RuntimeError("Expected Power BI token to passed as a command line argument.")

else:
    handler = PowerBIAPIClient(config["tenant_id"], config["application_id"], sys.argv[1])
    prod_workspace = config["production"]["name"]

    report_paths = list(
        map(
            lambda p: os.path.join(POWERBI_REPORTS_FOLDER, p),
            fnmatch.filter(os.listdir(POWERBI_REPORTS_FOLDER), "*.pbix"),
        )
    )
    dataset_paths = list(
        map(
            lambda p: os.path.join(POWERBI_DATASET_FOLDER, p),
            fnmatch.filter(os.listdir(POWERBI_DATASET_FOLDER), "*.pbix"),
        )
    )

    print("\nUploading data sets ...")
    for path in dataset_paths:
        display_name = get_display_name(path)
        handler.import_file_into_workspace(prod_workspace, True, path, display_name)
        print(f"Uploaded {path} as {display_name}")
    print("Done uploading data sets")

    print("\nUploading reports ...")
    for path in report_paths:
        display_name = get_display_name(path)
        handler.import_file_into_workspace(prod_workspace, False, path, display_name)
        print(f"Upladed {path} as {display_name}")
        if len(dataset_paths) == 1:
            dataset_name = get_display_name(dataset_paths[0])
            handler.rebind_report_in_workspace(prod_workspace, dataset_name, display_name)
            print(f"Rebound {display_name} to {dataset_name}")
    print("Done uploading reports")

    existing_datasets = set(map(lambda x: x["name"], handler.get_datasets_in_workspace(prod_workspace)))
    existing_reports = set(map(lambda x: x["name"], handler.get_reports_in_workspace(prod_workspace)))
    datasets_to_remove = existing_datasets - set([get_display_name(file) for file in dataset_paths])
    reports_to_remove = existing_reports - set([get_display_name(file) for file in report_paths])

    print("\nDeleting reports that should not be present ...")
    for path in reports_to_remove:
        handler.delete_report(prod_workspace, path)
        print("Deleted", path)
    print("Done deleting reports")

    print("\nDeleting dataset that should not be present ...")
    for path in datasets_to_remove:
        handler.delete_dataset(prod_workspace, path)
        print("Deleted", path)
    print("Done deleting data sets")



import glob
import logging
import os
import posixpath
from typing import Callable, Dict, List

from databricks_api import DatabricksAPI
from requests import HTTPError

import forge_template.util.file_util as file_util
import forge_template.util.log_utils as log_utils
from forge_template.handler.handler import FORBIDDEN_CODE, NOT_FOUND_CODE, BaseHandler
from forge_template.util.paths import DATABRICKS_PATH, DATABRICKS_TEMPLATE_PATH, notebooks_to_copy
from forge_template.util.yaml_util import YAML_CONFIG_PATH

CONFIG_PATH_PLACEHOLDER = "YAML_CONFIG_PATH_PLACEHOLDER"
DEFAULT_SCOPE_ACCESS = "WRITE"

WORKSPACE = "Workspace"
WORKSPACE_DIRECTORY = "Workspace Directory"
DBFS_DIRECTORY = "DBFS Directory"


class DatabricksHandler(BaseHandler):
    def __init__(self, name: str, config: Dict) -> None:
        config = config[name]
        self.client = DatabricksAPI(host=config["host"], token=config["token"])

        self._dev_workspace = config["development"]
        self._prod_workspace = config["production"]
        self._dbfs_path = config["config_dbfs_path"]

        self.workspaces_to_create: List[str] = []
        self.groups_to_create: List[str] = []
        self.scopes_to_create: List[Dict[str, str]] = []
        self.dbfs_dirs_to_create: List[str] = []
        self.users_to_add_to_group: Dict[str, List[str]] = {}
        self.secrets_to_add_to_scope: Dict[str, List[Dict[str, str]]] = {}
        self.groups_to_add_to_scope: Dict[str, List[str]] = {}

        self.created_workspaces: List[str] = []
        self.created_groups: List[str] = []
        self.created_scopes: List[str] = []
        self.created_dbfs_dirs: List[str] = []
        self.added_users_to_group: Dict[str, List[str]] = {}
        self.added_groups_to_scope: Dict[str, List[str]] = {}

        self.dbfs_config_path = ""

        super().__init__(name=name, config=config)

    def create_preview(self) -> None:
        self.check_dbfs_dir_to_add(self._dbfs_path)
        self.check_workspaces_to_add([self._dev_workspace["workspace_path"], self._prod_workspace["workspace_path"]])
        self.check_scopes_to_add([self._dev_workspace, self._prod_workspace])
        self.check_groups_to_add([self._dev_workspace["group"], self._prod_workspace["group"]])

    def print_preview(self):
        log_utils.print_resource_to_add(self.workspaces_to_create, "Workspaces")
        log_utils.print_resource_to_add(self.groups_to_create, "Groups")
        log_utils.print_resource_to_add(list(map(lambda scope: scope["scope_name"], self.scopes_to_create)), "Scopes")
        log_utils.print_resource_to_add(self.dbfs_dirs_to_create, "DBFS Directories")
        logging.info("")
        log_utils.print_resources_to_update(self.users_to_add_to_group, "User", "Group")
        log_utils.print_resources_to_update(self.secrets_to_add_to_scope, "Secret", "Scope")
        log_utils.print_resources_to_update(self.groups_to_add_to_scope, "Group", "Scope")

    def setup(self):
        self.create_workspaces(self.workspaces_to_create)
        self.create_dbfs_directories(self.dbfs_dirs_to_create)
        self.create_groups(self.groups_to_create)
        self.add_users_to_group(self.users_to_add_to_group)
        self.create_scopes(self.scopes_to_create)
        self.add_groups_to_scope(self.groups_to_add_to_scope)
        self.add_secrets_to_scope(self.secrets_to_add_to_scope)

    def rollback(self) -> None:
        if self.created_scopes:
            logging.info("Rolling back all scopes that have been created.")
            self.remove_scopes(self.created_scopes)
        if self.created_groups:
            logging.info("Rolling back all groups that have been added.")
            self.remove_groups(self.created_groups)
        if self.created_workspaces:
            logging.info("Rolling back all workspaces that have been created.")
            self.remove_workspace_directories(self.created_workspaces)
        if self.created_dbfs_dirs:
            logging.info("Rolling back all DBFS directories that have been added.")
            self.remove_dbfs_directories(self.created_dbfs_dirs)
        if self.added_users_to_group:
            groups_with_new_users = self.get_dict_list_subtraction(self.added_users_to_group, self.created_groups)
            if groups_with_new_users:
                logging.info("Rolling back all users that have been added to groups.")
                self.remove_users_from_groups(groups_with_new_users)
        if self.added_groups_to_scope:
            scopes_with_new_groups = self.get_dict_list_subtraction(self.added_groups_to_scope, self.created_scopes)
            if scopes_with_new_groups:
                logging.info("Rolling back all groups that have been added to scopes.")
                self.remove_groups_from_scope(scopes_with_new_groups)

    def delete_all_resources(self) -> None:
        self.remove_groups([self.config["production"]["group"]["name"], self.config["development"]["group"]["name"]])
        self.remove_dbfs_directories([self.config["config_dbfs_path"]])
        self.remove_scopes([self.config["production"]["scope"]["name"], self.config["development"]["scope"]["name"]])
        self.remove_workspace_directories(
            [self.config["production"]["workspace_path"], self.config["development"]["workspace_path"]]
        )

    def upload_config(self) -> None:
        self.check_dbfs_dir_to_add(self._dbfs_path)
        self.create_dbfs_directories(self.dbfs_dirs_to_create)
        self.copy_files_to_dbfs(self.dbfs_config_path, YAML_CONFIG_PATH)

    def check_groups_to_add(self, groups: List) -> None:
        """
        Checks which Databricks groups to add, as well as which users to add to the groups

        Args:
            groups: List of groups. See config.yaml for example.
        """
        existing_groups = self.get_existing_groups()
        for group in groups:
            group_name = group["name"]
            users = group["users"]

            if group_name in existing_groups:
                log_utils.print_resource_already_exists(group_name, "Group")
                self.check_users_to_add(group_name, users, True)
            else:
                self.check_users_to_add(group_name, users, False)
                self.groups_to_create.append(group_name)

    def check_users_to_add(self, group_name: str, users: List[str], group_exist: bool) -> None:
        if group_exist:
            current_users = self.get_existing_users_in_group(group_name)
            users = list(set(users) - set(current_users))

        self.users_to_add_to_group[group_name] = users

    def check_scopes_to_add(self, configs: List[Dict]) -> None:
        """
        Checks which Databricks scopes to add

        Args:
            configs: List of scope configs. See config.yaml and secrets.yaml for examples.
        """
        existing_scopes = self.get_existing_scopes()
        for config in configs:
            self.check_scope_to_add(config["scope"], existing_scopes, config["group"]["name"])

    def check_scope_to_add(self, scope_config: Dict, existing_scopes: List[str], group_name: str) -> None:
        """
        Checks whether or not to add a given scope, as well as which secrets to add to the scope

        Args:
            scope_config: Scope configuration. See config.yaml for example.
            existing_scopes: List with names of existing scopes
            group_name: Name of group that should have access to scope
        """
        scope_name = scope_config["name"]
        initial_manage_principal = (
            self.config["initial_manage_principal"]
            if (self.config["initial_manage_principal"] != "" and self.config["initial_manage_principal"] is not None)
            else None
        )

        scope_exists = False
        if scope_name in existing_scopes:
            log_utils.print_resource_already_exists(scope_name, "Scope")
            scope_exists = True
        else:
            self.scopes_to_create.append(
                {"scope_name": scope_name, "initial_manage_principal": initial_manage_principal}
            )

        self.check_group_to_add_to_scope(scope_name, group_name, scope_exists)
        self.check_secrets_to_add_to_scope(
            scope_name, scope_config["secrets"], scope_config["cdf_api_key"] if "cdf_api_key" in scope_config else ""
        )

    def check_group_to_add_to_scope(self, scope_name: str, group_name: str, scope_exists: bool) -> None:
        if scope_name not in self.groups_to_add_to_scope:
            self.groups_to_add_to_scope[scope_name] = []

        if scope_exists:
            if not self.does_acls_exist_in_scope(scope_name, group_name):
                self.groups_to_add_to_scope[scope_name].append(group_name)
        else:
            self.groups_to_add_to_scope[scope_name].append(group_name)

    def check_secrets_to_add_to_scope(self, scope_name: str, secrets: List[Dict[str, str]], cdf_api_key: str = ""):
        if cdf_api_key:
            secrets.append({"key": "cdf_api_key", "string_value": cdf_api_key})

        if scope_name not in self.secrets_to_add_to_scope:
            self.secrets_to_add_to_scope[scope_name] = []
        self.secrets_to_add_to_scope[scope_name].extend(secrets)

    def check_workspaces_to_add(self, workspace_paths: List) -> None:
        """
        Checks if the given workspace paths already exists.

        Args:
            workspace_paths: Workspace paths to check
        """
        for workspace_path in workspace_paths:
            if self.does_path_exist(workspace_path, lambda client, path: client.workspace.list(path)):
                log_utils.print_resource_already_exists(workspace_path, WORKSPACE)
            else:
                self.workspaces_to_create.append(workspace_path)

    def check_dbfs_dir_to_add(self, dbfs_path: str) -> None:
        """
        Checks if the given dbfs path already exists. If not, it will be added to a the list of dbfs paths to add

        Args:
            dbfs_path: Path to check
        """
        if self.does_path_exist(dbfs_path, lambda client, path: client.dbfs.list(path)):
            log_utils.print_resource_already_exists(dbfs_path, DBFS_DIRECTORY)
        else:
            self.dbfs_dirs_to_create.append(dbfs_path)

        self.dbfs_config_path = dbfs_path

    def create_groups(self, group_names: List[str]) -> None:
        """
        Creates Databricks groups

        Args:
            group_names: Names of groups to create
        """
        for name in group_names:
            self.client.groups.create_group(name)
            self.created_groups.append(name)
            log_utils.print_success_created(entity_name=name, entity_type="Group")

    def add_users_to_group(self, group_to_users: Dict[str, List[str]]):
        """
        Adds users to a specified group

        Args:
            group_to_users: Group name to list of user names
        """
        for group_name, users in group_to_users.items():
            for name in users:
                self.client.groups.add_to_group(group_name, user_name=name)
                if group_name not in self.added_users_to_group:
                    self.added_users_to_group[group_name] = []
                self.added_users_to_group[group_name].append(name)
                log_utils.print_success_added(group_name, "Group", name, "User")

    def create_scopes(self, scopes: List[Dict]) -> None:
        """
        Create Databricks scopes

        Args:
            scopes: List with scope names
        """
        for scope in scopes:
            self.client.secret.create_scope(
                scope=scope["scope_name"], initial_manage_principal=scope["initial_manage_principal"]
            )

            self.created_scopes.append(scope["scope_name"])
            log_utils.print_success_created(scope["scope_name"], "Scope")

    def add_secrets_to_scope(self, scope_to_secrets: Dict[str, List[Dict]]) -> None:
        """
        Add secrets to a specified scope

        Args:
            scope_to_secrets: Scope name to list of secrets where each secret is a dictionary with keys:
                ``"key"``
                Key/Identifier of secret (str)

                ``"string_value"``
                The actual secret value (str)
        """
        for scope, secrets in scope_to_secrets.items():
            for secret in secrets:
                self.client.secret.put_secret(scope, secret["key"], secret["string_value"])
                log_utils.print_success_added(scope, "Scope", secret["key"], "Token")

    # TODO: Handling of Databricks account that isn't premium
    def add_groups_to_scope(self, scope_to_group: Dict[str, List[str]]) -> None:
        """
        Add groups to a specified scope.

        Args:
            scope_to_group: Scope name to list of group names
        """
        for scope, groups in scope_to_group.items():
            for group in groups:
                try:
                    self.client.secret.put_acl(scope, group, DEFAULT_SCOPE_ACCESS)
                    if scope not in self.added_groups_to_scope:
                        self.added_groups_to_scope[scope] = []
                    self.added_groups_to_scope[scope].append(group)
                    log_utils.print_success_added(scope, "Scope", group, "Group")
                except HTTPError as e:
                    if e.response.status_code == FORBIDDEN_CODE:
                        logging.warning(f"Account doesn't support ACL.")
                        continue
                    raise e

    def create_workspaces(self, workspace_paths: List) -> None:
        """
        Creates Databricks workspaces and add notebook templates to the workspaces

        Args:
            workspace_paths: Paths to add workspace to in Databricks
        """
        file_util.copy_folders(DATABRICKS_TEMPLATE_PATH, DATABRICKS_PATH, ignore=["*template*", "__init__.py"])
        dbfs_location = f"/dbfs/{self.dbfs_config_path[1:]}/{YAML_CONFIG_PATH}"
        file_util.copy_files_and_replace_placeholders(notebooks_to_copy, CONFIG_PATH_PLACEHOLDER, dbfs_location)
        for path in workspace_paths:
            self.create_workspace(path)

    def create_workspace(self, base_path: str) -> None:
        """
        Creates a Databricks workspace and add notebook templates to the workspace

        Args:
            base_path: Path to add workspace to in Databricks
        """
        self.create_workspace_directory(base_path)
        self.created_workspaces.append(base_path)

        for folder_name in os.listdir(DATABRICKS_PATH):
            workspace_path = posixpath.join(base_path, folder_name)
            self.create_workspace_directory(workspace_path)
        self.copy_notebooks_to_workspace(base_path)

    def create_workspace_directory(self, directory_path: str) -> None:
        self.client.workspace.mkdirs(directory_path)
        log_utils.print_success_created(directory_path, WORKSPACE_DIRECTORY)

    def copy_notebooks_to_workspace(self, workspace_path: str) -> None:
        notebook_paths = glob.glob(os.path.join(DATABRICKS_PATH, "**", "*.py"))
        for path in notebook_paths:
            content = file_util.get_base64_encoded_content(path)
            path = path.replace(DATABRICKS_PATH, "")
            upload_path = file_util.get_absolute_posix_path(workspace_path, path)
            self.client.workspace.import_workspace(upload_path, language="PYTHON", format="SOURCE", content=content)

            notebook_name = os.path.basename(path)
            dir_name = os.path.dirname(upload_path)
            log_utils.print_success_added(dir_name, WORKSPACE_DIRECTORY, notebook_name, "Notebook")

    def copy_files_to_dbfs(self, dbfs_path: str, local_path: str) -> None:
        file_name = os.path.basename(local_path)
        content = file_util.get_base64_encoded_content(local_path)
        upload_path = file_util.get_absolute_posix_path(dbfs_path, file_name)
        self.client.dbfs.put(upload_path, contents=content)
        log_utils.print_success_added(dbfs_path, DBFS_DIRECTORY, file_name, "File")

    def create_dbfs_directories(self, directory_paths: List[str]) -> None:
        """
        Creates Databricks DBFS directories and copies the config.yaml file to those directories

        Args:
            directory_paths: List of DBFS paths to create
        """
        for path in directory_paths:
            self.client.dbfs.mkdirs(path)
            self.created_dbfs_dirs.append(path)
            log_utils.print_success_created(path, DBFS_DIRECTORY)
            self.copy_files_to_dbfs(path, YAML_CONFIG_PATH)

    def remove_groups(self, groups: List) -> None:
        self.remove_entities(
            groups, "Group", self.get_existing_groups, lambda client, group: client.groups.remove_group(group)
        )

    def remove_users_from_groups(self, group_to_users: Dict) -> None:
        """
        Removes users from a Databricks group

        Args:
            group_to_users: Group name to list of user names
        """
        self.remove_sub_entity_from_entity(
            group_to_users,
            self.get_existing_groups,
            self.get_existing_users_in_group,
            lambda client, group, user: client.groups.remove_from_group(group, user),
            "Group",
            "User",
        )

    def remove_scopes(self, scopes: List) -> None:
        self.remove_entities(
            scopes, "Scope", self.get_existing_scopes, lambda client, scope: client.secret.delete_scope(scope)
        )

    def remove_groups_from_scope(self, scope_to_groups: Dict) -> None:
        """
        Remove group permissions to scope in Databricks

        Args:
            scope_to_groups: Name of scope to list of group names
        """
        self.remove_sub_entity_from_entity(
            scope_to_groups,
            self.get_existing_scopes,
            self.get_existing_acls_in_scope,
            lambda client, scope, group: client.secret.delete_acl(scope, group),
            "Scope",
            "Group",
        )

    def remove_workspace_directories(self, directory_paths: List) -> None:
        self.remove_paths(
            directory_paths,
            lambda client, path: client.workspace.list(path),
            lambda client, path: client.workspace.delete(path, recursive=True),
            WORKSPACE_DIRECTORY,
        )

    def remove_dbfs_directories(self, directory_paths: List) -> None:
        self.remove_paths(
            directory_paths,
            lambda client, path: client.dbfs.list(path),
            lambda client, path: client.dbfs.delete(path, recursive=True),
            DBFS_DIRECTORY,
        )

    def remove_entities(
        self,
        entities: List[str],
        entity_type: str,
        get_existing: Callable[[List[str]], List[str]],
        remove: Callable[[DatabricksAPI, str], None],
    ):
        entities = get_existing(entities)
        for entity in entities:
            remove(self.client, entity)
            log_utils.print_success_deleted(entity, entity_type)

    def remove_sub_entity_from_entity(
        self,
        entity_to_sub: Dict[str, List[str]],
        get_existing_entities: Callable[[List[str]], List[str]],
        get_exisiting_sub_entities: Callable[[str, List[str]], List[str]],
        remover: Callable[[DatabricksAPI, str, str], None],
        entity_type: str,
        sub_entity_type: str,
    ) -> None:
        existing_entities = get_existing_entities(list(entity_to_sub.keys()))
        for entity in existing_entities:
            sub_entities = get_exisiting_sub_entities(entity, entity_to_sub[entity])
            for sub_entity in sub_entities:
                remover(self.client, entity, sub_entity)
                log_utils.print_success_removed(sub_entity, entity, sub_entity_type, entity_type)

    def remove_paths(
        self,
        paths: List[str],
        existence_checker: Callable[[DatabricksAPI, str], str],
        delete_method: Callable[[DatabricksAPI, str], None],
        path_type: str,
    ) -> None:
        for path in paths:
            if self.does_path_exist(path, existence_checker):
                delete_method(self.client, path)
                log_utils.print_success_deleted(path, path_type)

    def does_path_exist(self, dir_path: str, check_function: Callable) -> bool:
        """
        Check if a directory path already exists in Databricks

        Args:
            dir_path: Path of directory to check
            check_function: Function doing the checking
        """
        try:
            check_function(self.client, dir_path)
            return True
        except HTTPError as e:
            if e.response.status_code == NOT_FOUND_CODE:
                return False
            raise e

    def entity_exist(
        self, entity: str, getter: Callable[[DatabricksAPI], Dict], key: str, dict_to_list: Callable[[Dict], str] = None
    ) -> bool:
        return entity in self.get_existing_entities(getter, key, dict_to_list)

    def get_existing_entities(
        self,
        getter: Callable[[DatabricksAPI], Dict],
        key: str,
        dict_to_list: Callable[[Dict], str] = None,
        entities_to_check=None,
    ) -> List[str]:
        existing_entities_response = getter(self.client)
        existing_entities = self.get_entity_response_to_list(existing_entities_response, key, dict_to_list)
        if entities_to_check is None:
            return existing_entities
        else:
            return list(set(existing_entities).intersection(set(entities_to_check)))

    def get_existing_groups(self, groups: List[str] = None):
        return self.get_existing_entities(
            getter=lambda client: client.groups.get_groups(), key="group_names", entities_to_check=groups
        )

    def get_existing_scopes(self, scopes: List[str] = None):
        return self.get_existing_entities(
            getter=lambda client: client.secret.list_scopes(),
            key="scopes",
            entities_to_check=scopes,
            dict_to_list=lambda d: d["name"],
        )

    def get_existing_users_in_group(self, group_name: str, users: List[str] = None) -> List[str]:
        return self.get_existing_entities(
            getter=lambda client: client.groups.get_group_members(group_name),
            key="members",
            entities_to_check=users,
            dict_to_list=lambda d: d["user_name"],
        )

    def does_acls_exist_in_scope(self, scope_name: str, acl: str) -> bool:
        return acl in self.get_existing_acls_in_scope(scope_name)

    def get_existing_acls_in_scope(self, scope_name: str, acls: List[str] = None) -> List[str]:
        return self.get_existing_entities(
            getter=lambda client: client.secret.list_acls(scope_name),
            key="items",
            entities_to_check=acls,
            dict_to_list=lambda d: d["principal"],
        )

    @staticmethod
    def get_entity_response_to_list(entity_response: Dict, key: str, dict_transformer: Callable[[Dict], str] = None):
        if key not in entity_response:
            return []
        if dict_transformer:
            return [dict_transformer(entity) for entity in entity_response[key]]
        return entity_response[key]



import os
import shutil
from typing import Dict, List

from pbiapi import PowerBIAPIClient

import forge_template.util.log_utils as log_utils
from forge_template.handler.handler import BaseHandler
from forge_template.util.paths import (
    POWERBI_BASE_DATASET_PATH,
    POWERBI_BASE_REPORT_PATH,
    POWERBI_PATH,
    POWERBI_TEMPLATE_PATH,
)
from forge_template.util.yaml_util import assert_is_email


class PowerBIHandler(BaseHandler):
    def __init__(self, name: str, config: Dict) -> None:
        config = config[name]
        self._client = PowerBIAPIClient(config["tenant_id"], config["application_id"], config["token"])

        self._dev_workspace = config["development"]
        self._prod_workspace = config["production"]

        self._workspaces_to_create: List[str] = []
        self._users_to_add_to_workspace: Dict[str, List[str]] = {}
        self._users_to_update_in_workspace: Dict[str, List[Dict]] = {}

        self._added_workspaces: List[str] = []
        self._added_users_to_workspace: Dict[str, List[str]] = {}

        super().__init__(name=name, config=config)

    def create_preview(self) -> None:
        self.check_workspaces_to_add([self._dev_workspace, self._prod_workspace])

    def print_preview(self) -> None:
        if not self._workspaces_to_create and not self._users_to_add_to_workspace:
            log_utils.print_message("No changes to deploy")
        log_utils.print_resource_to_add(self._workspaces_to_create, "Workspace")
        log_utils.print_resources_to_update(self._users_to_add_to_workspace, "Principal", "Workspace")

    def setup(self) -> None:
        self.create_workspaces(self._workspaces_to_create)
        self.add_users_to_workspace(self._users_to_add_to_workspace)

    def rollback(self) -> None:
        if self._added_workspaces:
            log_utils.print_rollback("workspaces")
            self.__remove_workspaces(self._added_workspaces)

    def delete_all_resources(self):
        self.__remove_workspaces([self._dev_workspace["name"], self._prod_workspace["name"]])

    def check_workspaces_to_add(self, workspaces: List) -> None:
        """
        Checks which PowerBI workspaces to add, as well as which users to add to the workspace.

        Args:
            workspaces: List of workspace definitions. See config.yaml for example.
        """
        existing_workspaces = self.__get_existing_workspaces()
        for workspace in workspaces:
            workspace_name = workspace["name"]
            if workspace_name not in existing_workspaces.keys():
                self._workspaces_to_create.append(workspace_name)
                self._users_to_add_to_workspace[workspace_name] = workspace["principals"]

    def __get_existing_workspaces(self) -> Dict:
        response = self._client.get_workspaces()
        return {w["name"]: w for w in response}

    def __remove_workspaces(self, workspace_names: List) -> None:
        for workspace_name in workspace_names:
            self._client.delete_workspace(workspace_name)
            log_utils.print_success_deleted(workspace_name, "Workspace")

    def create_workspaces(self, workspace_names: List) -> None:
        """
        Creates workspaces in PowerBI

        Args:
            workspace_names: Names of workspaces to create
        """
        self.__copy_folders_locally()
        for name in workspace_names:
            self.create_workspace(name)

    def create_workspace(self, workspace_name: str) -> None:
        """
        Creates a workspace in PowerBI and uploads a dataset to that workspace.

        Args:
            workspace_name: Name of workspace to create
        """
        self._client.create_workspace(workspace_name)
        self._added_workspaces.append(workspace_name)
        self.__upload_files_to_workspace(workspace_name)
        self.__rebind_report_to_dataset(workspace_name)
        log_utils.print_success_created(workspace_name, "Workspace")

    def add_users_to_workspace(self, workspace_name_to_users: Dict[str, List[str]]) -> None:
        """
        Add a users to an existing PowerBI workspace.

        Args:
            workspace_name_to_users: Workspace name to list of users where each user is a dictionary with keys:
                ``"identifier"``
                Id of user to add (email, group id or app id) (str)

                ``"access_right"``
                Permission level to give to the user (Admin, Member, Contributor) (str)
        """
        for workspace_name, users in workspace_name_to_users.items():
            for user in users:
                self.add_user_to_workspace(user, workspace_name)

    def add_user_to_workspace(self, user_id: str, workspace_name: str, access_right: str = "Admin") -> None:
        """
        Add a user to an existing workspace and set the permissions of that user

        Args:
            user_id: Id of user. A user can be a person, group or Power App.
            workspace_name: Name of workspace where the user will be added
            access_right: Permission level to give to the user
        """
        user = {"identifier": user_id, "groupUserAccessRight": access_right}
        if assert_is_email(user_id):
            user["UserEmailAddress"] = user_id
        self._client.add_user_to_workspace(workspace_name, user)
        if workspace_name not in self._added_users_to_workspace.keys():
            self._added_users_to_workspace[workspace_name] = []
        self._added_users_to_workspace[workspace_name].append(user_id)
        log_utils.print_success_added(workspace_name, "Workspace", user_id, "Principal")

    def __upload_files_to_workspace(self, workspace_name: str) -> None:
        pbix_paths = [POWERBI_BASE_DATASET_PATH, POWERBI_BASE_REPORT_PATH]
        for file_path in pbix_paths:
            display_name = self.get_display_name_from_pbix_path(file_path)
            skip_report = "dataset" in display_name
            self.__upload_file_to_workspace(workspace_name, file_path, display_name, skip_report)

    def __upload_file_to_workspace(
        self, workspace_name: str, local_path: str, display_name: str, skip_report: bool = False
    ) -> None:
        self._client.import_file_into_workspace(workspace_name, skip_report, local_path, display_name)
        log_utils.print_success_added(workspace_name, "Workspace", display_name, "PBIX file")

    def __rebind_report_to_dataset(self, workspace_name: str) -> None:
        dataset_name = self.get_display_name_from_pbix_path(POWERBI_BASE_DATASET_PATH)
        report_name = self.get_display_name_from_pbix_path(POWERBI_BASE_REPORT_PATH)

        self._client.rebind_report_in_workspace(workspace_name, dataset_name, report_name)
        self._client.delete_dataset(workspace_name, report_name)

    @staticmethod
    def __copy_folders_locally() -> None:
        if not os.path.exists(POWERBI_PATH):
            shutil.copytree(POWERBI_TEMPLATE_PATH, POWERBI_PATH)

    @staticmethod
    def get_display_name_from_pbix_path(path: str) -> str:
        return os.path.splitext(os.path.basename(path))[0]



class AccessDeniedException(Exception):
    """ Raise when a invalid access token is being used """

    pass


class GitHandlerException(Exception):
    """ Raise when GitHandler fails """

    pass


class ValidationException(Exception):
    """ Raise when Cerberus validator fails """

    pass



import base64
import os
import posixpath
import re
import shutil
from typing import List, Tuple, Union


def copy_folders(src: str, dest: str, ignore: Union[List[str], str] = "", overwrite: bool = False) -> None:
    if overwrite or not os.path.exists(dest):
        if ignore:
            if isinstance(ignore, str):
                ignore = [ignore]
            shutil.copytree(src, dest, ignore=shutil.ignore_patterns(*ignore))
        else:
            shutil.copytree(src, dest)


def prepare_rel_posix_path(path: str) -> str:
    """
        Prepare any path in posix format and remove first slash if present
    """
    split_path = re.split(r"\\{1,}|/{1,}", path)
    remove_empty_path = list(filter(None, split_path))
    posix_path = "/".join(remove_empty_path)

    return posix_path


def get_absolute_posix_path(prefix: str, suffix: str = "") -> str:
    prefix = prepare_rel_posix_path(prefix)
    prefix = posixpath.join(posixpath.sep, prefix)
    suffix = prepare_rel_posix_path(suffix)

    return posixpath.join(prefix, suffix)


def copy_files_and_replace_placeholders(src_dest_pairs: List[Tuple[str, str]], placeholder: str, replacement: str):
    for src, dest in src_dest_pairs:
        with open(src, "r") as f:
            content = f.read()

        content = content.replace(placeholder, replacement)
        with open(dest, "w") as f:
            f.write(content)


def get_base64_encoded_content(file_path: str) -> str:
    with open(file_path, "rb") as f:
        content = f.read()
        encoded_content = base64.b64encode(content).decode()

        return encoded_content



import logging
from typing import List

from git import Repo

import forge_template.util.log_utils as log_utils

FORGE_REMOTE_NAME = "forge-template"


def push_content_to_repo(repo: Repo, repo_name: str) -> None:
    repo.index.add("*")
    added_files = [key[0] for key in repo.index.entries.keys()]
    repo.git.commit("-m", "initial commit")
    repo.git.push("--set-upstream", "origin", "master")
    for added_file in added_files:
        log_utils.print_success_added(repo_name, "Repository", added_file, "File")


def check_remotes_to_add(repo: Repo, remote_name: str, remotes_to_rename: List, remotes_to_add: List) -> None:
    origin_exists = is_remote_in_remotes(repo, "origin")
    if origin_exists:
        remotes_to_rename.append(["origin", FORGE_REMOTE_NAME])
        remotes_to_add.append(remote_name)


def add_remotes(remotes_to_add: List, url: str, repo: Repo, remotes_added: List) -> None:
    for remote in remotes_to_add:
        repo.create_remote(remote, url=url)
        remotes_added.append(remote)
        log_utils.print_success_created(remote, "Remote")


def rename_remotes(remotes_to_rename: List, repo: Repo, renamed_remotes: List) -> List:
    for (old_name, new_name) in remotes_to_rename:
        if is_remote_in_remotes(repo, old_name):
            repo.remote(old_name).rename(new_name)
            renamed_remotes.append((new_name, old_name))
            logging.info(f"Renamed remote {old_name} to {new_name}")
    return renamed_remotes


def remove_remotes(remote_list: List, repo: Repo) -> None:
    for remote in remote_list:
        if is_remote_in_remotes(repo, remote):
            repo.git.remote("rm", remote)
            log_utils.print_success_deleted(remote, "Remote")


def is_remote_in_remotes(repo: Repo, remote_name: str) -> bool:
    return remote_name in [remote.name for remote in repo.remotes]



# Databricks notebook source
# MAGIC %md # Setup development environment
# MAGIC
# MAGIC ### Prerequisites
# MAGIC Python 3.7+
# MAGIC
# MAGIC ### About this notebook
# MAGIC This notebook initializes the development environment. This notebook should be run in the top of all other notebooks. It assumes that a Databricks widget called _notebook_path_ exists. The notebook will to the following for you:
# MAGIC - **Cmd 2**: Install three python libraries; ruamel.yaml (for reading the .yaml config file), pbiapi (Python SDK for Power BI), congite-sdk.
# MAGIC - **Cmd 3**: Figure out if this notebook is in the development or production environment. If running this notebook from a job, the parameter 'stage' needs to be set.
# MAGIC - **Cmd 4**: Read the config file and store the data in the file as a Python dictionary in a variable named _config_
# MAGIC - **Cmd 5**: Initializes a Cognite Python SDK client and stores it in the variable named _client_
# MAGIC
# MAGIC To use in other notebooks, simply include a cell with the following code: `%run ../setup/setup-environment.ipynb`. Please note that the file path is relative to the notebook you run it from.

# COMMAND ----------

# Install necessary Python libraries on the cluster
# ruamel.yaml is used to read the config file
# pbiapi is used for interacting with PowerBI (triggering a dataset refresh)
dbutils.library.installPyPI("ruamel.yaml")
dbutils.library.installPyPI("pbiapi", version="0.1.4")
dbutils.library.installPyPI("cognite-sdk", version="1.5.0")

# COMMAND ----------

# Which stage (development or production) is this notebook?
try:
  dbutils.notebook.entry_point.getDbutils().notebook().getContext().currentRunId().get()
  stage = dbutils.widgets.get("stage")
except Exception:
  path = dbutils.notebook.entry_point.getDbutils().notebook().getContext().notebookPath().get()
  if "/prod/" in path:
    stage = "production"
  else:
    stage = "development"

# COMMAND ----------

from cognite.client import CogniteClient
from ruamel.yaml import YAML

if stage not in ["production", "development"]:
  raise InputError(f"Illegal stage, {stage}")

# Read config file
yaml_config_path = "YAML_CONFIG_PATH_PLACEHOLDER"
with open(yaml_config_path, encoding="utf-8") as f:
    full_config = YAML(typ="safe").load(f)

if "powerbi" in full_config:
    powerbi_config = full_config["powerbi"]  # This config is only needed if working with PowerBI
db_config = full_config["databricks"]
config = db_config[stage]
config["project_description"] = db_config["project_description"]
config["group_name"] = config["group"]["name"]
config["api_key"] = dbutils.secrets.get(
    config["scope"]["name"], "cdf_api_key"
)  # Assumes that the cdf-read-key exists in the scope
print("Done reading config. Available as a Python variable named 'config' (a dictionary)")


# COMMAND ----------

api_key = config["api_key"]
project = config["cdf_project_name"]
client_name = f"forge-{config['cdf_project_name']}"
client = CogniteClient(api_key=api_key, project=project, client_name=client_name)
print("Done creating Cognite Python SDK client. Available as a Python variables named 'client'")



# Databricks notebook source
# MAGIC %md # Setup base model (database and views)
# MAGIC
# MAGIC ### Prerequisites
# MAGIC Python 3.7+
# MAGIC
# MAGIC ### About this notebook
# MAGIC This notebook create a database with views. There is one view for each of the following CDF resource types: assets, events, datapoints, timeseries, files, 3dmodels.
# MAGIC
# MAGIC - **Cmd 2**: The setup-environment notebook is run to setup the development environment (installing commonly used libraries and reading a configuration file). The parameters from the configuration file is available as a Python dictionary named _config_.

# COMMAND ----------

# MAGIC %run ../setup/setup-environment.py

# COMMAND ----------

# Create database:
spark.sql(
    """
    CREATE DATABASE IF NOT EXISTS {database_name} 
    COMMENT '{project_description}'
    """.format(
        **config
    )
)

# COMMAND ----------

# Grant permissions:
spark.sql("GRANT ALL PRIVILEGES ON DATABASE {database_name} TO `{group_name}`".format(**config))

# COMMAND ----------

# Create table with assets
spark.sql("DROP TABLE IF EXISTS {database_name}.api_assets".format(**config))
spark.sql(
    """
CREATE TABLE IF NOT EXISTS {database_name}.api_assets USING cognite.spark.v1
OPTIONS (type 'assets',
         apiKey '{api_key}',
         partitions '80')
COMMENT 'View of assets in CDF from base model'
""".format(
        **config
    )
)

# COMMAND ----------

# Create table with events
spark.sql("DROP TABLE IF EXISTS {database_name}.api_events".format(**config))
spark.sql(
    """
  CREATE TABLE IF NOT EXISTS {database_name}.api_events USING cognite.spark.v1
  OPTIONS (type 'events',
           apiKey '{api_key}',
           partitions '80')
  COMMENT 'View of events in CDF from base model'
  """.format(
        **config
    )
)

# COMMAND ----------

# Create table with datapoints
spark.sql("DROP TABLE IF EXISTS {database_name}.api_datapoints".format(**config))
spark.sql(
    """
  CREATE TABLE IF NOT EXISTS {database_name}.api_datapoints USING cognite.spark.v1
  OPTIONS (type 'datapoints',
           apiKey '{api_key}')
  COMMENT 'View of datapoints in CDF from base model'
  """.format(
        **config
    )
)

# COMMAND ----------

# Create table with timeseries
spark.sql("DROP TABLE IF EXISTS {database_name}.api_timeseries".format(**config))
spark.sql(
    """
  CREATE TABLE IF NOT EXISTS {database_name}.api_timeseries USING cognite.spark.v1
  OPTIONS (type 'timeseries',
           apiKey '{api_key}')
  COMMENT 'View of time series in CDF from base model'
  """.format(
        **config
    )
)

# COMMAND ----------

# Create table with files metadata
spark.sql("DROP TABLE IF EXISTS {database_name}.api_files".format(**config))
spark.sql(
    """
  CREATE TABLE IF NOT EXISTS {database_name}.api_files USING cognite.spark.v1
  OPTIONS (type 'files',
           apiKey '{api_key}')
  COMMENT 'View of files metadata in CDF from base model'
  """.format(
        **config
    )
)

# COMMAND ----------

# Create table with 3D models and revisions metadata
spark.sql("DROP TABLE IF EXISTS {database_name}.api_3dmodels".format(**config))
spark.sql(
    """
  CREATE TABLE IF NOT EXISTS {database_name}.api_3dmodels USING cognite.spark.v1
  OPTIONS (type '3dmodels',
           apiKey '{api_key}')
  COMMENT 'View of 3D models and revisions metadata in CDF from base model'
  """.format(
        **config
    )
)

# COMMAND ----------

# Display database table metadata
display(spark.catalog.listTables(config["database_name"]))

# COMMAND ----------

# Display schemas for database tables
for t in ["api_assets", "api_datapoints", "api_events", "api_timeseries", "api_files", "api_3dmodels"]:
    print(">>", t)
    spark.table("%s.%s" % (config["database_name"], t)).printSchema()



import logging
from typing import Any, Dict, List


def print_message(msg: str) -> None:
    logging.info(msg)


def print_entity_exists(entity_name: str = "", entity_type: str = "") -> None:
    logging.info(f"Databricks {entity_type} with name {entity_name} already exists.")


def print_success_created(entity_name: str = "", entity_type: str = "") -> None:
    logging.info(f"Successfully created {entity_name} [{entity_type}].")


def print_success_deleted(entity_name: str = "", entity_type: str = "") -> None:
    logging.info(f"Successfully deleted {entity_name} [{entity_type}].")


def print_success_removed(
    sub_entity_name: str = "", entity_name: str = "", sub_entity_type: str = "", entity_type: str = ""
) -> None:
    logging.info(f"Successfully removed {sub_entity_name} [{sub_entity_type}] from {entity_name} [{entity_type}].")


def print_success_added(parent_entity_name="", parent_entity_type="", entity_name="", entity_type="") -> None:
    logging.info(f"Successfully added {entity_name} [{entity_type}] to {parent_entity_name} " f"[{parent_entity_type}]")


def print_rollback(entity_type: str) -> None:
    logging.info(f"Rolling back all {entity_type} that have been created")


def print_resources_to_update(resource_dict: Dict, resource_type_to_add: str, main_resource_type: str) -> None:
    for resource_name, resource_list in resource_dict.items():
        if resource_list:
            print_resource_update(resource_list, resource_name, resource_type_to_add, main_resource_type)


def print_remotes_to_rename(rename_list: List) -> None:
    for (original_name, new_name) in rename_list:
        logging.info(f"Remote: {original_name} will be renamed to {new_name}")


def print_files_to_be_added() -> None:
    logging.info("All files in this repository except those in .gitignore will be added to the new remote")


def print_resource_update(
    resource_list: List, main_resource: str, resource_type_to_add: str, main_resource_type: str
) -> None:
    logging.info(
        f"{main_resource_type}: "
        + ", ".join([f"{resource} [{resource_type_to_add}]" for resource in resource_list])
        + f" will be added to {main_resource} [{main_resource_type}]"
    )


def print_resource_to_add(resource_list: List[Any], resource_type: str) -> None:
    if resource_list:
        for resource in resource_list:
            logging.info(f"{resource_type}: {resource} will be added.")


def print_resource_already_exists(resource_name: str, resource_type: str, override: bool = False) -> None:
    logging.info(
        f"Resource {resource_name} [{resource_type}] already exists. It will "
        + ("" if override else "not")
        + " be re-created."
    )



import os
import shutil
import tempfile
from os import listdir
from os.path import isfile, join
from typing import Dict, List

from azure.devops.connection import Connection
from azure.devops.exceptions import AzureDevOpsServiceError
from azure.devops.v5_1.build.models import (
    AgentPoolQueue,
    BuildDefinition,
    BuildDefinitionVariable,
    BuildRepository,
    TeamProjectReference,
)
from azure.devops.v5_1.git.models import GitRepositoryCreateOptions
from git import Repo
from msrest.authentication import BasicAuthentication

import forge_template.handler.git_utils as git_utils
import forge_template.util.log_utils as print_utils
from forge_template.handler.handler import BaseHandler
from forge_template.util.yaml_util import DATABRICKS_CONFIG_NAME, PIPELINES_PATH, POWERBI_CONFIG_NAME

PIPELINE_SERVICES = ["databricks", "powerbi"]


class AzureHandler(BaseHandler):
    """
    Handler for interacting with Azure DevOps.
    Currently supporting:
        - Create Azure Repo
        - Create Azure Build Definition
    """

    def __init__(self, name: str, config: Dict):
        self._databricks_config = config[DATABRICKS_CONFIG_NAME]
        self._powerbi_config = config[POWERBI_CONFIG_NAME]

        config = config[name]
        self._organization = config["principal"]
        self._project_name = config["project_name"]
        self._repo_name = config["repo_name"]

        url = f"https://dev.azure.com/{self._organization}"
        credentials = BasicAuthentication("", config["token"])
        connection = Connection(base_url=url, creds=credentials)
        self._core_client = connection.clients.get_core_client()
        self._build_client = connection.clients.get_build_client()
        self._git_client = connection.clients.get_git_client()

        self._git_tmp_work_dir = tempfile.TemporaryDirectory().name
        self._local_repo = Repo.init(path=self._git_tmp_work_dir)

        self._project_id: str = ""
        self._repo_url: str = ""
        self._repo_ssh_url: str = ""
        self._definition_ids: Dict[str, str] = {}

        self._build_definitions_to_create: List[_BuildDefinition] = []
        self._created_build_definitions: List[Dict[str, str]] = []

        self._repos_to_create: List[str] = []
        self._remotes_to_rename: List[str] = []
        self._remotes_to_add: List[str] = []

        self._created_repos: Dict[str, str] = {}
        self._renamed_remotes: List[str] = []
        self._added_remotes: List[str] = []

        super().__init__(name=name, config=config)

    @property
    def __project_id(self) -> str:
        return self._project_id if self._project_id else self.__get_project_id()

    def __get_project_id(self) -> str:
        project = self._core_client.get_project(project_id=self._project_name)
        self._project_id = project.id
        return project.id

    @property
    def __repo_url(self) -> str:
        return self._repo_url if self._repo_url else self.__get_repository_urls()

    @property
    def __repo_ssh_url(self) -> str:
        return self._repo_ssh_url if self._repo_ssh_url else self.__get_repository_urls(return_ssh=True)

    @property
    def __repo_id(self) -> str:
        return self.__get_repository_id()

    @property
    def __definition_ids(self) -> Dict[str, str]:
        return self._definition_ids if self._definition_ids else self.__get_existing_definitions(self.__project_id)

    def __get_repository_urls(self, return_ssh: bool = False) -> str:
        try:
            repository = self._git_client.get_repository(repository_id=self._repo_name, project=self.__project_id)
            self._repo_ssh_url = repository.ssh_url
            self._repo_url = repository.url
            return repository.ssh_url if return_ssh else repository.url
        except AzureDevOpsServiceError as e:
            if e.type_key == "GitRepositoryNotFoundException":
                return ""
            raise e

    def __get_repository_id(self):
        repository = self._git_client.get_repository(repository_id=self._repo_name, project=self.__project_id)
        return repository.id

    def create_preview(self) -> None:
        if not self.__repo_ssh_url:
            self._repos_to_create.append(self._repo_name)
        git_utils.check_remotes_to_add(self._local_repo, self._repo_name, self._remotes_to_rename, self._remotes_to_add)
        self.create_definition_preview(self.__project_id)

    def print_preview(self) -> None:
        print_utils.print_resource_to_add([self._repo_name], "Repository")
        print_utils.print_remotes_to_rename(self._remotes_to_rename)
        print_utils.print_resource_to_add(self._remotes_to_add, "Remote")
        print_utils.print_resource_to_add(self._build_definitions_to_create, "Build definition")

    def setup(self) -> None:
        for repo_name in self._repos_to_create:
            self.create_repo(repo_name, self.__project_id)

        try:
            remote_name = self._local_repo.remote()
        except ValueError:
            remote_name = None
        if self._repo_ssh_url is not None and remote_name is None:
            self._local_repo.create_remote("origin", self._repo_ssh_url)

        self.__copy_files_to_repo_dir()

        git_utils.push_content_to_repo(self._local_repo, self._repo_name)

        for definition in self._build_definitions_to_create:
            self.create_definition(
                name=definition.name, variables=definition.build_variables, file_path=definition.file_path
            )

    def rollback(self) -> None:
        for definition in self._created_build_definitions:
            self.delete_definition(definition=definition)

        for repo_id, repo_name in self._created_repos.items():
            self.delete_repo(name=repo_name, repo_id=repo_id)

        git_utils.remove_remotes(self._added_remotes, self._local_repo)
        git_utils.rename_remotes(self._renamed_remotes, self._local_repo, [])

    def delete_all_resources(self):
        self.delete_repo(self._repo_name, self.__repo_id)
        for def_name, def_id in self.__definition_ids.items():
            self.delete_definition({"name": def_name, "id": def_id})

    def create_repo(self, repo_name: str, project_id: str) -> None:
        """
        Creates a new Azure DevOps Repository

        Args:
            repo_name: Name of repository
            project_id: Id of Azure DevOps project where the repository will be created
        """
        create_repo = GitRepositoryCreateOptions(name=repo_name)
        response = self._git_client.create_repository(create_repo, project=project_id)
        self._created_repos[repo_name] = response.id
        self._repo_ssh_url = response.ssh_url
        print_utils.print_success_created(repo_name, "Repository")

    def delete_repo(self, name: str, repo_id: str) -> None:
        """
        Deletes an Azure DevOps Repository

        Args:
            name: Name of repository
            repo_id: Id of Azure DevOps Repository that will be deleted
        """
        self._git_client.delete_repository(repo_id, project=self._project_id)
        print_utils.print_success_deleted(name, "Repository")

    def create_definition_preview(self, project_id: str):
        """
        Checks which build definitions that does not exist. All definition files in the pipelines folder are checked.
        If a build definition does not exist, it is added to the list of definitions (_build_definitions_to_create)
        that are to be created.
        """
        existing_definitions = self.__get_existing_definitions(project_id=project_id)
        for file_path in os.listdir(PIPELINES_PATH):
            service = os.path.splitext(file_path)[0]
            definition_name = f"{self._project_name}-{service}"
            if definition_name not in existing_definitions:
                self._build_definitions_to_create.append(
                    _BuildDefinition(
                        definition_name, self.__get_build_variables(service), os.path.join(PIPELINES_PATH, file_path),
                    )
                )

    def create_definition(self, name: str, variables: Dict[str, BuildDefinitionVariable], file_path: str,) -> None:
        """
        Creates a Azure DevOps Build definition

        Args:
            name: Name of definition
            variables: Dictionary mapping variable name to its definition (BuildDefinitionVariable)
            file_path: Path of .yaml/.yml file containing the definition
        """

        build_repo = BuildRepository(
            default_branch="refs/heads/master",
            checkout_submodules=False,
            name=self._repo_name,
            url=self.__repo_url,
            type="TfsGit",
            properties={"reportBuildStatus": True},
        )

        queue = AgentPoolQueue(name="Azure Pipelines")
        triggers = {"settingsSourceType": 2, "batchChanges": False, "triggerType": "continuousIntegration"}
        project = TeamProjectReference(id=self.__project_id)
        process = {"type": 2, "yamlFilename": file_path}

        definition = BuildDefinition(
            name=name,
            project=project,
            repository=build_repo,
            process=process,
            type="build",
            variables=variables,
            queue=queue,
            triggers=[triggers],
        )

        response = self._build_client.create_definition(definition, self.__project_id)
        self._created_build_definitions.append({"name": name, "id": response.id})
        print_utils.print_success_created(name, "Build definition")

    def delete_definition(self, definition: Dict[str, str]) -> None:
        """
        Deletes a definition

        Args:
            definition: Definition to be deleted as a dictionary with keys:
                ``"id"``
                Id of definition to delete (str)

                ``"name"``
                Name of definition to delete (str)
        """
        self._build_client.delete_definition(self.__project_id, definition["id"])
        print_utils.print_success_deleted(definition["name"], "Build definition")

    def __get_build_variables(self, service_name: str) -> Dict[str, BuildDefinitionVariable]:
        if service_name == "databricks":
            return {
                "DatabricksHost": BuildDefinitionVariable(value=self._databricks_config["host"]),
                "DatabricksToken": BuildDefinitionVariable(value=self._databricks_config["token"], is_secret=True),
                "DatabricksNotebookFolder": BuildDefinitionVariable(value="databricks"),
                "DatabricksWorkspaceFolder": BuildDefinitionVariable(
                    value=self._databricks_config["production"]["workspace_path"]
                ),
            }
        elif service_name == "powerbi":
            return {
                "POWERBI_TENANT_ID": BuildDefinitionVariable(value=self._powerbi_config["tenant_id"]),
                "POWERBI_APPLICATION_ID": BuildDefinitionVariable(value=self._powerbi_config["application_id"]),
                "POWERBI_TOKEN": BuildDefinitionVariable(value=self._powerbi_config["token"], is_secret=True),
                "POWERBI_WORKSPACE": BuildDefinitionVariable(value=self._powerbi_config["production"]["name"]),
            }
        else:
            raise RuntimeError(f"Unexpected service name. Got {service_name}, but expected on of [databricks, powerbi]")

    def __get_existing_definitions(self, project_id: str) -> Dict[str, str]:
        """
        Fetches the existing build definitions for the current project id

        Returns: A list of definition names
        """
        definitions = self._build_client.get_definitions(project_id).value
        return {definition.name: definition.id for definition in definitions}

    def __copy_files_to_repo_dir(self) -> None:
        working_dir = os.getcwd()

        pipelines_path = os.path.join(working_dir, "pipelines")
        template_databricks_path = os.path.join(working_dir, "template", "databricks")
        template_powerbi_path = os.path.join(working_dir, "template", "powerbi")
        template_infra_path = os.path.join(working_dir, "template", "infra_scripts")
        config_path = os.path.join(working_dir, "config.yaml")
        requirements_path = os.path.join(working_dir, "requirements.txt")

        self.__handle_copy_dir_files(pipelines_path, "pipelines")
        self.__handle_copy_dir_files(template_databricks_path, "databricks")
        self.__handle_copy_dir_files(template_powerbi_path, "powerbi")
        self.__handle_copy_dir_files(template_infra_path, "infra_scripts")

        shutil.copy(config_path, self._git_tmp_work_dir)
        shutil.copy(requirements_path, self._git_tmp_work_dir)

    def __handle_copy_dir_files(self, source_dir: str, file_type: str) -> None:
        files = [
            join(source_dir, f) for f in listdir(source_dir) if isfile(join(source_dir, f)) and f != "." and f != ".."
        ]
        dest_path = os.path.join(self._git_tmp_work_dir, file_type)
        os.mkdir(dest_path)
        for file in files:
            shutil.copy(file, dest_path)


class _BuildDefinition:
    def __init__(self, name: str, build_variables: Dict[str, BuildDefinitionVariable], file_path: str):
        self.name = name
        self.build_variables = build_variables
        self.file_path = file_path



# Databricks notebook source
# MAGIC %md # Spark SQL sample
# MAGIC
# MAGIC ### About this notebook
# MAGIC This notebook create very simple model using Spark SQL. It creates two views, one for timeseries and one for datapoints associated with those time series

# COMMAND ----------

# MAGIC %run ../setup/setup-environment.py

# COMMAND ----------

# Creating a sample table with timeseries
ts = spark.sql(
    """    
  CREATE OR REPLACE VIEW {database_name}.timeseries
  COMMENT 'Timeseries used for base Power BI model' AS 
  SELECT 
    name,
    id,
    unit
  FROM {database_name}.api_timeseries
  LIMIT 10
""".format(
        **config
    )
)

# COMMAND ----------

# Select ids of all timeseries related to pumps
ts_ids = ",".join(spark.sql("SELECT id FROM {database_name}.timeseries".format(**config)).toPandas()["id"].astype(str))

# COMMAND ----------

# Creating a sample table with datapoints
dp = spark.sql(
    """
  CREATE OR REPLACE VIEW {database_name}.datapoints AS
  SELECT 
    id,
    timestamp,
    value
  FROM
    {database_name}.api_datapoints
  WHERE
    timestamp > TIMESTAMP(ADD_MONTHS(NOW(),-36)) AND
    timestamp < TIMESTAMP(DATE_ADD(NOW(),1)) AND
    aggregation = 'average' AND
    granularity = '1d' AND
    id in ({ts_ids})
  """.format(
        **config, ts_ids=ts_ids
    )
)



import os

TEMPLATES_PATH = os.path.join(os.path.dirname(__file__), "..", "templates")

# Databricks constants
DATABRICKS_TEMPLATE_PATH = os.path.join(TEMPLATES_PATH, "databricks" + os.sep)
DATABRICKS_PATH = "databricks" + os.sep
notebooks_to_copy = [
    (
        os.path.join(DATABRICKS_TEMPLATE_PATH, "setup", "setup-environment-template.py"),
        os.path.join(DATABRICKS_PATH, "setup", "setup-environment.py"),
    ),
    (
        os.path.join(DATABRICKS_TEMPLATE_PATH, "setup", "basemodel-template.py"),
        os.path.join(DATABRICKS_PATH, "setup", "basemodel.py"),
    ),
    (
        os.path.join(DATABRICKS_TEMPLATE_PATH, "jobs", "powerbi-dataset-refresh-template.py"),
        os.path.join(DATABRICKS_PATH, "jobs", "powerbi-dataset-refresh.py"),
    ),
    (
        os.path.join(DATABRICKS_TEMPLATE_PATH, "notebooks", "sample-spark-model-template.py"),
        os.path.join(DATABRICKS_PATH, "notebooks", "sample-spark-model-template.py"),
    ),
]

# Power BI constants
POWERBI_TEMPLATE_PATH = os.path.join(TEMPLATES_PATH, "powerbi" + os.sep)
POWERBI_PATH = "powerbi" + os.sep
POWERBI_BASE_DATASET_PATH = os.path.join(POWERBI_PATH, "datasets", "base_dataset.pbix")
POWERBI_BASE_REPORT_PATH = os.path.join(POWERBI_PATH, "reports", "base_report.pbix")

# Github Actions constants
SCRIPT_TEMPLATE_FOLDER = os.path.join(TEMPLATES_PATH, "infra_scripts")
WORKFLOW_TEMPLATE_FOLDER = os.path.join(TEMPLATES_PATH, "github_workflows")
POWERBI_SCRIPT_NAME = "deploy_powerbi.py"
DATABRICKS_SCRIPT_NAME = "deploy_databricks.py"
POWERBI_SCRIPT_PATH = os.path.join(SCRIPT_TEMPLATE_FOLDER, POWERBI_SCRIPT_NAME)
DATABRICKS_SCRIPT_PATH = os.path.join(SCRIPT_TEMPLATE_FOLDER, DATABRICKS_SCRIPT_NAME)
POWERBI_WORKFLOW_TEMPLATE_NAME = "deploy_powerbi.yaml"
DATABRICKS_WORKFLOW_TEMPLATE_NAME = "deploy_databricks.yaml"
POWERBI_WORKFLOW_TEMPLATE_PATH = os.path.join(WORKFLOW_TEMPLATE_FOLDER, POWERBI_WORKFLOW_TEMPLATE_NAME)
DATABRICKS_WORKFLOW_TEMPLATE_PATH = os.path.join(WORKFLOW_TEMPLATE_FOLDER, DATABRICKS_WORKFLOW_TEMPLATE_NAME)



import logging
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List

from requests import HTTPError

from forge_template.exception.exceptions import AccessDeniedException

NOT_FOUND_CODE = 404
UNAUTHORIZED_CODE = 401
FORBIDDEN_CODE = 403
BAD_REQUEST_CODE = 400


class BaseHandler(ABC):
    def __init__(self, name: str, config: Dict[str, Any]):
        self.name = name
        self.capitalized_name = name.capitalize()
        self.config = config

    def __run_function(self, f: Callable, text: str, print_footer: bool = False, should_rollback: bool = False) -> None:
        header = f"{text}"
        logging.info(header)
        try:
            f(self)
        except HTTPError as e:
            status = e.response.status_code
            if status in [UNAUTHORIZED_CODE, FORBIDDEN_CODE]:
                raise AccessDeniedException(
                    f"Access denied. Please check your {self.capitalized_name} access token. Details: {e}"
                )
            self.__handle_rollback(should_rollback=should_rollback)
            raise
        except Exception:
            self.__handle_rollback(should_rollback=should_rollback)
            raise

        if print_footer:
            footer = "-" * len(header)
            logging.info(footer)

    def __handle_rollback(self, should_rollback: bool) -> None:
        if should_rollback:
            self.rollback()

    def run_setup(self) -> None:
        text = f"\nCreating/Updating resources for {self.capitalized_name} ..."
        self.__run_function(lambda handler: handler.setup(), text, should_rollback=True)

    @abstractmethod
    def setup(self) -> None:
        pass

    def run_create_preview(self) -> None:
        text = f"\nCreating preview of resource creation/updates for {self.capitalized_name} ..."
        self.__run_function(lambda handler: handler.create_preview(), text)

    @abstractmethod
    def create_preview(self) -> None:
        pass

    def run_print_preview(self) -> None:
        text = f"\n{self.name} preview".upper()
        self.__run_function(lambda handler: handler.print_preview(), text)

    @abstractmethod
    def print_preview(self) -> None:
        pass

    def run_rollback(self) -> None:
        text = f"\nRolling back {self.capitalized_name} due to creation/update error"
        self.__run_function(lambda handler: handler.rollback(), text)

    @abstractmethod
    def rollback(self) -> None:
        pass

    def perform_action(self, action: Callable, should_rollback: bool = True) -> Any:
        try:
            return action(self)
        except HTTPError as e:
            if e.response.status_code == FORBIDDEN_CODE:
                raise AccessDeniedException(f"Access denied. Please check your {self.capitalized_name} access token.")
            if should_rollback:
                self.rollback()
            raise e
        except RuntimeError as e:
            if should_rollback:
                self.rollback()
            raise e

    def run_delete_all_resources(self):
        text = f"\nDeleting all resources for {self.capitalized_name} ..."
        self.__run_function(lambda handler: handler.delete_all_resources(), text)

    @abstractmethod
    def delete_all_resources(self):
        pass

    @staticmethod
    def get_dict_list_subtraction(d: Dict, l: List) -> Dict:
        return {e: d[e] for e in set(d) - set(l)}



import base64
import os
import sys
from collections import deque

from databricks_api import DatabricksAPI
from ruamel.yaml import YAML

CONFIG_FILE_PATH = "config.yaml"

with open(CONFIG_FILE_PATH) as f:
    config = YAML(typ="safe").load(f)["databricks"]

if len(sys.argv) < 2:
    raise RuntimeError("Expected Databricks token to passed as a command line argument.")

else:
    handler = DatabricksAPI(host=config["host"], token=sys.argv[1])
    prod_workspace = config["production"]["workspace_path"]
    stack = deque(["databricks"])
    local_paths = []
    while stack:
        curr_path = stack.pop()
        if os.path.isfile(curr_path):
            local_paths.append(curr_path)
        else:
            sub_paths = list(map(lambda p: os.path.join(curr_path, p), os.listdir(curr_path)))
            stack.extend(sub_paths)

    upload_paths = list(map(lambda p: p.replace("databricks", prod_workspace), local_paths))
    upload_dirs = set(map(lambda p: os.path.dirname(p).replace("databricks", prod_workspace), local_paths))

    print("\nDelete production workspace ...")
    handler.workspace.delete(prod_workspace, recursive=True)
    print("Done deleting production workspace")

    print("\nCreating directories in production workspace ...")
    for d in upload_dirs:
        handler.workspace.mkdirs(d)
        print(f"Created {d}")
    print("Done creating directories")

    print("\nUploading notebooks ...")
    for local_path, upload_path in zip(local_paths, upload_paths):
        with open(local_path, "rb") as f:
            content = f.read()
            encoded_content = base64.b64encode(content).decode()
            handler.workspace.import_workspace(upload_path, language="PYTHON", format="SOURCE", content=encoded_content)
            print(f"Uploaded {local_path} as {upload_path}")
    print("Done uploading notebooks")



import collections
import os
import re
from typing import Dict, Optional, Tuple

from cerberus import Validator
from ruamel.yaml import YAML

from forge_template.exception.exceptions import ValidationException

SCHEMA_FOLDER = os.path.join(os.path.dirname(__file__), "..", "schema")
YAML_CONFIG_PATH = "config.yaml"
YAML_SECRETS_PATH = "secrets.yaml"
YAML_CONFIG_SCHEMA_PATH = os.path.join(SCHEMA_FOLDER, "config_schema.yaml")
YAML_SECRETS_SCHEMA_PATH = os.path.join(SCHEMA_FOLDER, "secrets_schema.yaml")

PIPELINES_PATH = "pipelines"

DATABRICKS_CONFIG_NAME = "databricks"
GITHUB_CONFIG_NAME = "github"
AZURE_CONFIG_NAME = "azure"
POWERBI_CONFIG_NAME = "powerbi"


def load_yaml(path: str) -> dict:
    yaml_reader = YAML(typ="safe")

    with open(path, encoding="utf-8") as f:
        dct = yaml_reader.load(f)
        assert isinstance(dct, dict), f"YAML malformatted, couldn't load into dict (got {type(dct)})"
        return dct


def validate_yaml(schema_path: str, document: Dict) -> Tuple[bool, Optional[Dict]]:
    schema = load_yaml(schema_path)
    validator = Validator(schema)
    return validator.validate(document), validator.errors or None


def load_and_validate_yaml(path: str, schema_path: str) -> Dict:
    loaded_yaml = load_yaml(path)
    is_valid, errors = validate_yaml(schema_path, loaded_yaml)
    if not is_valid:
        raise ValidationException(f"Validation of {path} failed.\n{errors}")
    return loaded_yaml


def assert_same_services(config: Dict, secrets: Dict) -> None:
    config_set = set(config.keys())
    secrets_set = set(secrets.keys())
    is_valid = config_set == secrets_set
    diff = {}
    if len(config_set - secrets_set) != 0:
        diff["secret.yaml is missing"] = config_set - secrets_set
    if len(secrets_set - config_set) != 0:
        diff["config.yaml is missing"] = secrets_set - config_set

    assert is_valid, f"YAML files don't contain the same services defined. Differences:\n{diff}"


def merge_dictionaries(dict_a: Dict, dict_b: Dict) -> None:
    for k, v in dict_b.items():
        if k in dict_a and isinstance(dict_a[k], dict) and isinstance(dict_b[k], collections.Mapping):
            merge_dictionaries(dict_a[k], dict_b[k])
        else:
            dict_a[k] = dict_b[k]


def assert_is_email(s: str) -> bool:
    regex = re.compile(r"^\w+([.-]?\w+)*@\w+([.-]?\w+)*(\.\w{2,3})+$")
    return re.search(regex, s) is not None


def assert_scope_matches(config: Dict, secrets: Dict) -> None:
    if "databricks" in config and "databricks" in secrets:
        for stage in ["production", "development"]:
            name_in_config = config["databricks"][stage]["scope"]["name"]
            name_in_secrets = secrets["databricks"][stage]["scope"]["name"]
            assert name_in_config == name_in_secrets is not None, (
                f"Mismatch between scope name in '{YAML_CONFIG_PATH}' ({name_in_config}) and "
                f"'{YAML_SECRETS_PATH}' ({name_in_secrets})."
            )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['forge_template',
 'forge_template.exception',
 'forge_template.handler',
 'forge_template.templates.databricks.jobs',
 'forge_template.templates.databricks.notebooks',
 'forge_template.templates.databricks.setup',
 'forge_template.templates.infra_scripts',
 'forge_template.util']

package_data = \
{'': ['*'],
 'forge_template': ['schema/*',
                    'templates/github_workflows/*',
                    'templates/pipelines/*',
                    'templates/powerbi/datasets/*',
                    'templates/powerbi/reports/*']}

install_requires = \
['azure-devops>=6.0.0b2,<7.0.0',
 'cerberus>=1.3,<2.0',
 'click>=7.0,<8.0',
 'colorama==0.4.1',
 'databricks-api>=0.3.0,<0.4.0',
 'gitpython>=3.1.0,<4.0.0',
 'pbiapi>=0.1.4,<0.2.0',
 'pyfiglet>=0.8.post1,<0.9',
 'requests>=2.22,<3.0',
 'ruamel.yaml>=0.16.5,<0.17.0',
 'termcolor>=1.1.0,<2.0.0']

entry_points = \
{'console_scripts': ['forge-template = forge_template.main:cli']}

setup_kwargs = {
    'name': 'cognite-forge-template',
    'version': '0.5.6',
    'description': 'Automation of software used for data science in Cognite, including Databricks, PowerBI, Git and Grafana',
    'long_description': None,
    'author': 'Magnus Moan',
    'author_email': 'magnus.moan@cognite.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# Databricks notebook source
# MAGIC %md
# MAGIC # Power BI dataset refresh
# MAGIC In order for the dataset refresh to work a PowerBI service principal token needs to be present in the scope as _powerbi-key_.
# MAGIC Furthermore, for the refresh to work, the correct server url, database name and databricks credentials needs to be set on the Power BI side. This can be done by following these steps:
# MAGIC 1. Log into Power BI
# MAGIC 2. Navigate to your workspace
# MAGIC 3. Go to Datasets
# MAGIC 4. In the Action column click the three dots next to the base_dataset.
# MAGIC 5. Choose settings
# MAGIC 6. Under Datasource Credentials and Parameters fill in the necessary information.

# COMMAND ----------

# MAGIC %run ../setup/setup-environment.py

# COMMAND ----------

import requests
# PowerBI: Trigger a dataset refresh
from pbiapi import PowerBIAPIClient

if not powerbi_config:
    print(
        "Power BI not configured in the config.yaml file. Please overwrite the file manually with a version containing Power BI configuration, or use the forge-template cli to do it."
    )
else:
    tenant_id = powerbi_config["tenant_id"]
    client_id = powerbi_config["application_id"]
    client_secret = dbutils.secrets.get(config["scope"]["name"], "powerbi-key")
    workspace_name = powerbi_config[stage]["name"]

    client = PowerBIAPIClient(tenant_id=tenant_id, client_id=client_id, client_secret=client_secret)
    datasets = client.get_datasets_in_workspace(workspace_name)
    for dataset in datasets:
        client.refresh_dataset_by_id(workspace_name, dataset["id"])



from typing import Callable, Dict, List, Optional

import colorama
from click import confirm
from pyfiglet import Figlet
from termcolor import colored, cprint


class UserInterface:
    colorama.init()

    @staticmethod
    def print_title(title):
        cprint(Figlet(font="slant").renderText(title), "white")

    @staticmethod
    def print_failure(msg):
        cprint(msg, "red")

    @staticmethod
    def print_warning(msg):
        cprint(msg, "yellow")

    @staticmethod
    def print_ok(msg):
        cprint(msg, "green")

    @staticmethod
    def print_in_progress(msg):
        cprint(msg, "blue")

    def get_input(
        self,
        prompt: str,
        default: str = "",
        choices: Optional[List[str]] = None,
        validate: Optional[Callable[[str], bool]] = None,
    ) -> str:
        _prompt = self.format_prompt(prompt, choices, default)
        value = input(_prompt) or default
        if choices and value not in choices:
            self.print_failure("Must be one of {}".format(choices))
            return self.get_input(prompt, default, choices, validate)
        if validate and validate(value) is False:
            return self.get_input(prompt, default=default, choices=choices, validate=validate)
        return value

    def get_menu_selection(self, prompt, choices: List[str], is_top_level: bool = False) -> str:
        print()
        for i, choice in enumerate(choices):
            print("[{}] {}".format(colored(str(i), "green"), choice))
        print()
        if not is_top_level:
            print("[{}] {}".format(colored("m", "green"), "To Main Menu"))
        print("[{}] {}".format(colored("e", "green"), "Exit"))
        print()
        choice = self.get_input(prompt)

        i_choices = [str(i) for i in range(len(choices))] + (["e"] if is_top_level else ["m", "e"])
        while choice not in i_choices:
            self.print_failure("Must be one of {}".format(i_choices))
            choice = self.get_input(prompt)

        if choice in ["e", "m"]:
            return choice
        return choices[int(choice)]

    def get_input_list_of_dicts(self, prompt: str, validators: Dict[str, Callable[[str], bool]], keys: List[str]):
        result = []
        while confirm(prompt, prompt_suffix="? ", show_default=True, default=False):
            values = {}
            for key in keys:
                values[key] = self.get_input(key.capitalize(), validate=validators[key])

            result.append(values)

        return result

    @staticmethod
    def __key_value_pair_to_dict(s: str, key: str, value: str, splitter: str) -> Dict[str, str]:
        elements = s.split(splitter)
        return {key: elements[0], value: elements[1]}

    @staticmethod
    def __valid_key_value_input(s: str) -> bool:
        values = s.split(",")
        for value in values:
            if ":" not in value:
                return False
        return True

    @staticmethod
    def format_prompt(prompt: str, choices: Optional[List[str]], default: str) -> str:
        _prompt = prompt
        if choices:
            _prompt += colored(" ({})".format("/".join(choices)), "blue")
        if default:
            _prompt += colored(" [{}]".format(default), "cyan")
        return _prompt + ": "
